/*
 *  Copyright (c) 2019 by Refinitiv. All rights reserved.
 *
 *  No portion of this software in any form may be used or
 *  reproduced in any manner without written consent from
 *  Refinitiv
 */
package com.refinitiv.analytics.reportingservice;

import java.io.Console;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Validate, Parse and hold command line arguments
 */
public class CmdLineHandler {

  private static final Logger logger = LoggerFactory.getLogger(CmdLineHandler.class);

  private static final List<String> reportTypesToSupport = Arrays.asList("RTS28", "FITCA", "RTS27");
  private static final String RUN = "RUN";
  private static final String QUERY = "QUERY";
  private static final String GET = "GET";
  private static final List<String> jobsToSupport = Arrays.asList(RUN, QUERY, GET);
  private static final String USER = "username";
  private static final String PWD = "password";
  private static final String START = "start";
  private static final String REPORT = "report";
  private static final String CLASS = "class";
  private static final String KEY = "key";
  private static final String END = "end";
  private static final String FIRM = "firm";
  private static final String DATA = "data";
  private static final String EMAIL = "email";

  private final Map<String, Options> mapOfOptions = new HashMap<>();
  private final Map<String, String> mapOfArgKeyValues = new HashMap<>();
  private String reportType;
  private String cmdJob;

  public CmdLineHandler() {
    Option userOption = new Option("u", USER, true, "AAA user name");
    Option pwdOption = new Option("p", PWD, true, "AAA password");
    Option jobOption = new Option("j", "job", true, "the job id");
    Option helpOption = new Option("h", "help", false, "show help message and exit");
    for (String action : jobsToSupport) {
      Options opt = new Options();
      if (action.equals(RUN)) {
        opt.addOption(userOption)
            .addOption(pwdOption)
            .addOption("d", DATA, true,
                "local data directory path which contain files of client raw data")
            .addOption("k", KEY, true, "the key used to encrypt data")
            .addOption("s", START, true, "start date, default <previousYear>-01-01")
            .addOption("e", END, true, "end date, default <previousYear>-12-31")
            .addOption("f", FIRM, true, "firm entity id, default ALL firms")
            .addOption("c", CLASS, true, "permId of CoFI, default ALL CoFIs")
            .addOption("m", EMAIL, true,
                "the email address for notification. Default to use the username")
            .addOption(helpOption);
      } else if (action.equals(GET)) {
        opt.addOption(userOption)
            .addOption(pwdOption)
            .addOption(jobOption)
            .addOption("r", REPORT, true, "local directory path to store reports downloaded")
            .addOption("k", KEY, true,
                "the key used to decrypt reports, must be same as to the key you used to run job")
            .addOption(helpOption);
      } else if (action.equals(QUERY)) {
        opt.addOption(userOption)
            .addOption(pwdOption)
            .addOption(jobOption)
            .addOption(helpOption);
      } else {
        continue;
      }

      this.mapOfOptions.put(action, opt);
    }
  }

  /**
   * Parse hand hold arguments from command line
   *
   * @param args String[]
   * @return boolean If parsing is successful
   * @throws Exception
   */
  public boolean parseArgs(String[] args) {
    if (!this.validateRequiredArgs(args)) {
      return false;
    }

    if (!this.parseOriginArgs(args)) {
      return false;
    }

    if (!this.allowUserInputCredentialArgs()) {
      return false;
    }

    if (this.cmdJob.equals(RUN)) {
      return this.parseRunJobArgs();
    } else if (this.cmdJob.equals(QUERY)) {
      return this.parseQueryJobArgs();
    } else if (this.cmdJob.equals(GET)) {
      return this.parseDownloadJobArgs();
    }

    return true;
  }

  private boolean validateRequiredArgs(String[] args) {
    if (args.length < 2) {
      this.printUsage("");
      return false;
    }

    this.reportType = args[0].toUpperCase();
    if (!reportTypesToSupport.contains(this.reportType)) {
      String errorString = String.format(
          "\"%s\" is not legal report type. Accepted values are: %s.%nType --help to view command help.",
          this.reportType, reportTypesToSupport.stream().collect(Collectors.toList()));
      logger.info(errorString);

      return false;
    }

    this.cmdJob = args[1].toUpperCase();
    if (!jobsToSupport.contains(this.cmdJob)) {
      String errorString = String.format(
          "\"%s\" is not legal Action type. Accepted values are: %s.%nType --help to view command help.",
          this.cmdJob, jobsToSupport.stream().collect(Collectors.toList()));
      logger.info(errorString);

      return false;
    }

    return true;
  }

  private boolean parseOriginArgs(String[] args) {
    this.mapOfArgKeyValues.clear();
    Options actionOpts = this.mapOfOptions.get(this.cmdJob);
    CommandLineParser parser = new DefaultParser();
    CommandLine cmd;
    try {
      cmd = parser.parse(actionOpts, args);
    } catch (ParseException e) {
      logger.info("Fail to parse the command line arguments.");
      logger.error(e.getMessage());
      return false;
    }

    // Print Help message if --help/-h option represents
    if (cmd.hasOption("h")) {
      this.printUsage(this.cmdJob);
      return false;
    }

    for (Option opt : actionOpts.getOptions()) {
      if (cmd.hasOption(opt.getOpt())) {
        this.mapOfArgKeyValues.put(opt.getLongOpt(), cmd.getOptionValue(opt.getLongOpt()));
      }
    }

    return true;
  }

  private boolean allowUserInputCredentialArgs() {
    if (!this.mapOfArgKeyValues.containsKey(USER)) {
      Console console = this.getCommandLineConsole();
      if (console == null) {
        return false;
      }
      String username = null;
      while (StringUtils.isBlank(username)) {
        username = console.readLine("Enter your AAA user name: ");
      }
      this.mapOfArgKeyValues.put(USER, username);
    }

    if (!this.mapOfArgKeyValues.containsKey(PWD)) {
      Console console = this.getCommandLineConsole();
      if (console == null) {
        return false;
      }
      String password = null;
      while (StringUtils.isBlank(password)) {
        char[] passwordArray = console.readPassword("Enter your AAA password: ");
        password = new String(passwordArray);
      }
      this.mapOfArgKeyValues.put(PWD, password);
    }

    return true;
  }

  private boolean parseRunJobArgs() {
    if (!this.mapOfArgKeyValues.containsKey(DATA)) {
      Console console = this.getCommandLineConsole();
      if (console == null) {
        return false;
      }
      String dataPath = null;
      while (StringUtils.isBlank(dataPath)) {
        dataPath = console.readLine("Enter your data directory for upload: ");
      }
      this.mapOfArgKeyValues.put(DATA, dataPath);
    }

    if (!this.mapOfArgKeyValues.containsKey(KEY)) {
      Console console = this.getCommandLineConsole();
      if (console == null) {
        return false;
      }
      String key = null;
      while (StringUtils.isBlank(key)) {
        key = console.readLine("Enter an key for data encryption: ");
      }
      this.mapOfArgKeyValues.put(KEY, key);
    }

    if (!this.mapOfArgKeyValues.containsKey(START)) {
      this.mapOfArgKeyValues.put(START, "");
    }

    if (!this.mapOfArgKeyValues.containsKey(END)) {
      this.mapOfArgKeyValues.put(END, "");
    }

    if (!this.mapOfArgKeyValues.containsKey(FIRM)) {
      this.mapOfArgKeyValues.put(FIRM, "");
    }

    if (!this.mapOfArgKeyValues.containsKey(CLASS)) {
      this.mapOfArgKeyValues.put(CLASS, "");
    }

    if (!this.mapOfArgKeyValues.containsKey(EMAIL)) {
      // use username as default notification email
      this.mapOfArgKeyValues.put(EMAIL, this.mapOfArgKeyValues.get(USER));
    }

    return true;
  }

  private boolean parseQueryJobArgs() {
    if (!this.mapOfArgKeyValues.containsKey("job")) {
      Console console = this.getCommandLineConsole();
      if (console == null) {
        return false;
      }
      String jobId = null;
      while (StringUtils.isBlank(jobId)) {
        jobId = console.readLine("Enter an job id for query: ");
      }
      this.mapOfArgKeyValues.put("job", jobId);
    }

    return true;
  }

  private boolean parseDownloadJobArgs() {
    if (!this.mapOfArgKeyValues.containsKey("job")) {
      Console console = this.getCommandLineConsole();
      if (console == null) {
        return false;
      }
      String jobId = null;
      while (StringUtils.isBlank(jobId)) {
        jobId = console.readLine("Enter an job id for downloading: ");
      }
      this.mapOfArgKeyValues.put("job", jobId);
    }

    if (!this.mapOfArgKeyValues.containsKey(REPORT)) {
      Console console = this.getCommandLineConsole();
      if (console == null) {
        return false;
      }
      String reportPath = null;
      while (StringUtils.isBlank(reportPath)) {
        reportPath = console.readLine("Enter a directory to store reports: ");
      }
      this.mapOfArgKeyValues.put(REPORT, reportPath);
    }

    if (!this.mapOfArgKeyValues.containsKey(KEY)) {
      Console console = this.getCommandLineConsole();
      if (console == null) {
        return false;
      }
      String key = null;
      while (StringUtils.isBlank(key)) {
        key = console.readLine("Enter an key for data decryption: ");
      }
      this.mapOfArgKeyValues.put(KEY, key);
    }

    return true;
  }

  private Console getCommandLineConsole() {
    Console console = System.console();
    if (console == null) {
      logger.info(
          "Couldn't get Console instance for user to input. Please make sure using correct command.");
    }
    return console;
  }

  private void printUsage(String action) {
    if (this.mapOfOptions.containsKey(action)) {
      String cmdLineSyntax = "java -jar MiFID_Reporting.jar " + action + " [...arguments...] [-h]";
      HelpFormatter formatter = new HelpFormatter();
      formatter.printHelp(cmdLineSyntax, this.mapOfOptions.get(action));
    } else {
      final String usageText =
          "usage: java -jar MiFID_Reporting.jar ReportType Action [...arguments...] [-h]\n"
              + "   ReportType     one of strings: "
              + reportTypesToSupport.stream().collect(Collectors.toList())
              + ", should be the first argument \n" + "   Action         one of strings: "
              + jobsToSupport.stream().collect(Collectors.toList())
              + ", should be the second argument\n"
              + "   -u username, --username username   AAA user name(used for non-interactive mode),\n"
              + "                                  if no user in the command, prompt to ask client type in AAA username follow the command\n"
              + "   -p password, --password  password   AAA password(used for non-interactive mode),\n"
              + "                                  if no password in the command, prompt to ask client type in AAA password follow the command\n"
              + "   -h, --help     show this help message and exit\n" + "   \n" + "For examples:\n"
              + "  java -jar MiFID_Reporting.jar RTS28 RUN -u \"username\" -p \"password\" -d \"dataPath\" -k \"encryptKey\" -s \"2018-01-01\" -e \"2018-12-31\" -f \"firmEntity\" -c \"CoFI\"\n"
              + "  java -jar MiFID_Reporting.jar RTS28 QUERY -u \"username\" -p \"password\" -j \"jobId\"\n"
              + "  java -jar MiFID_Reporting.jar RTS28 GET -u \"username\" -p \"password\" -j \"jobId\" -r \"reportPath\" -k \"encryptKey\"\n";
      logger.info(usageText);
    }
  }

  public String getReportType() {
    return this.reportType;
  }

  public String getCmdAction() {
    return this.cmdJob;
  }

  public String getOptionValue(String optionName) {
    return this.mapOfArgKeyValues.get(optionName);
  }
}
