/*
 *  Copyright (c) 2019 by Refinitiv. All rights reserved.
 *
 *  No portion of this software in any form may be used or
 *  reproduced in any manner without written consent from
 *  Refinitiv
 */
package com.refinitiv.analytics.reportingservice;

import com.amazonaws.auth.AWSCredentialsProvider;
import com.amazonaws.auth.AWSStaticCredentialsProvider;
import com.amazonaws.auth.BasicSessionCredentials;
import com.refinitiv.analytics.reportingservice.utils.EdpClient;
import com.refinitiv.analytics.reportingservice.utils.S3FileTransfer;
import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang3.StringUtils;
import org.json.JSONException;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Handler to run reporting jobs
 */
public class JobHandler {

  private static final Logger logger = LoggerFactory.getLogger(JobHandler.class);
  private static final String APPID = "ff12cd3431f746d0b60d6df351db1ad230d1ee75";
  private static final String EDP_ENDPOINT_4_CLOUD_CREDENTIAL = "205843-reporting-service-s3-apt";
  private static final String ERROR_FLAG = "error";

  private static final String MSG_FAIL_TO_RUN = "Fail to run Run Job due to bad paramters";
  private static final String MSG_CREDENTIALS_NULL = "Password should not be blank";

  private final EdpClient edpApiClient;

  public JobHandler() {
    this.edpApiClient = new EdpClient();
  }

  /**
   * Execute jobs from Command Line
   */
  public void execute(CmdLineHandler cmdLine) {
    String cmdAction = cmdLine.getCmdAction();
    try {
      if (cmdAction.equals("RUN")) {
        // do "RUN Job" job
        String username = cmdLine.getOptionValue("username");
        String password = cmdLine.getOptionValue("password");
        String localDir = cmdLine.getOptionValue("data");
        String encryptionKey = cmdLine.getOptionValue("key");
        String startTime = cmdLine.getOptionValue("start");
        String endTime = cmdLine.getOptionValue("end");
        String firmId = cmdLine.getOptionValue("firm");
        String cofi = cmdLine.getOptionValue("class");
        String email = cmdLine.getOptionValue("email");
        String reportType = cmdLine.getReportType();

        logger.info("Starting Run Job");
        this.executeRunJob(username, password, localDir, encryptionKey, startTime,
            endTime, firmId, cofi, email, reportType);
      } else if (cmdAction.equals("QUERY")) {
        // do "Query Job Status" job
        String username = cmdLine.getOptionValue("username");
        String password = cmdLine.getOptionValue("password");
        String reportType = cmdLine.getReportType();
        String jobId = cmdLine.getOptionValue("job");

        logger.info("Starting Query Job");
        String status = this.executeQueryJob(username, password, reportType, jobId).toString();
        logger.info("Job status: {}", status);
      } else if (cmdAction.equals("GET")) {
        // do Get Reports job
        String username = cmdLine.getOptionValue("username");
        String password = cmdLine.getOptionValue("password");
        String reportType = cmdLine.getReportType();
        String jobId = cmdLine.getOptionValue("job");
        String localDir = cmdLine.getOptionValue("report");
        String encryptionKey = cmdLine.getOptionValue("key");

        logger.info("Starting Get Report Job");
        if (this
            .executeGetReportsJob(username, password, reportType, jobId, localDir, encryptionKey)
            .has("success")) {
          logger.info("Reports downloaded.");
        } else {
          logger.error("Download failed");
        }
      } else {
        logger.info("Do not support action: {}", cmdAction);
      }
    } catch (Exception e) {
      String exceptionMsg = e.getMessage();
      logger.error(exceptionMsg);

      // parse HTTP request error and print its brief to user console
      String logMsg = "Fail to run action: " + cmdAction;
      Pattern p = Pattern.compile(".*Response.*, (code=\\d{3}, message=\\D*), url=.*");
      Matcher m = p.matcher(exceptionMsg);
      if (m.find()) {
        logMsg += " due to HTTP response failure: \"" + m.group(1) + "\"";
      }
      logger.info("{}. Find more information in the log file.", logMsg);
    }
  }

  /**
   * Execute RUN Job job
   *
   * @param username      String
   * @param password      String
   * @param localDir      String This is where the user's input located locally
   * @param encryptionKey String This is the key to encrypt user's input files
   * @param startTime     String
   * @param endTime       String
   * @param firmId        String
   * @param cofi          String
   * @param email         String
   * @param reportType    String
   * @return JSONObject {"jobId": "xxxxxxxxxxxxxxxxxxxxx"} on successful; {"error": "error
   * information"} on error
   * @throws IOException
   * @throws Exception
   */
  public JSONObject executeRunJob(String username, String password, String localDir,
      String encryptionKey, String startTime,
      String endTime, String firmId, String cofi, String email, String reportType)
      throws IOException {
    JSONObject resultJson = new JSONObject();
    if (!this.validateParamsForRunJob(username, password, localDir,
        encryptionKey, startTime, endTime, email, reportType)) {
      return resultJson.put(ERROR_FLAG, "Bad parameter(s)");
    }

    String token = this.getToken(username, password);

    String[] uploadS3Info = this.getS3InfoFromUploadLocationRequest(reportType, token);
    String bucketName = uploadS3Info[0];
    String region = uploadS3Info[1];
    String jobInfoFilePath = "";
    Pattern p = Pattern.compile("(.*/rept-srv-(.*))/input");
    Matcher m = p.matcher(uploadS3Info[2]);
    if (m.find()) {
      jobInfoFilePath = m.group(1);
    } else {
      throw new IllegalArgumentException(
          "Can not parse jobInfoFilePath and jobId from file upload path");
    }
    String[] credentials = this.getEdpClouldCredentials(token, region, bucketName, jobInfoFilePath);

    AWSCredentialsProvider credentialsProvider = new AWSStaticCredentialsProvider(
        new BasicSessionCredentials(credentials[0], credentials[1], credentials[2]));
    Date credentialTimestamp = new Date();

    String processedEncryptKey = this
        .generateKey(username, this.parseJobId(uploadS3Info[2]), encryptionKey);

    Integer filesUploaded = 0;
    File directory = new File(localDir);
    File[] files = directory.listFiles();
    logger.info("Start uploading file...");
    boolean isIncludeSchmaFile = false;
    for (int i = 0; i < files.length; i++) {
      long minuteElapsed = (new Date().getTime() - credentialTimestamp.getTime()) / 1000 / 60;
      if (minuteElapsed > 50) {
        logger.info("User credentials expired, re-request it.");
        credentialsProvider = this
            .getAWSCredentialsProviderWithEdpClouldCredentials(username, password, region,
                bucketName, jobInfoFilePath);
        credentialTimestamp = new Date();
      }
      if (files[i].isFile()) {
        String fileName = files[i].getName();
        if ("rts27.schema".equals(fileName) && "RTS27".equals(reportType)) {
          isIncludeSchmaFile = true;
        } else if (S3FileTransfer.shouldSkip(fileName)) {
          continue;
        }
        boolean uploadResultFlag = false;
        int retryTime = 3;
        while (!uploadResultFlag && retryTime > 0) {
          retryTime--;
          if (S3FileTransfer.uploadFile(files[i], uploadS3Info[0], uploadS3Info[1], uploadS3Info[2],
              processedEncryptKey, credentialsProvider)) {
            filesUploaded++;
            uploadResultFlag = true;
          }
        }
      }
    }

    if (filesUploaded > 1 || (!isIncludeSchmaFile && filesUploaded == 1)) {
      logger.info("Files uploaded");

      long minuteElapsed = (new Date().getTime() - credentialTimestamp.getTime()) / 1000 / 60;
      if (minuteElapsed > 50) {
        logger.info("User credentials expired, re-request it.");
        credentialsProvider = this
            .getAWSCredentialsProviderWithEdpClouldCredentials(username, password, region,
                bucketName, jobInfoFilePath);
        token = this.getToken(username, password);
      }
      JSONObject jsonObject = new JSONObject();
      try {
        jsonObject.put("bucket", uploadS3Info[0]);
        jsonObject.put("prefix", uploadS3Info[2]);
        jsonObject.put("cryptographyKey", processedEncryptKey);
        jsonObject.put("notification", email);
        if ("RTS28".equals(reportType)) {
          JSONObject filterObject = new JSONObject();
          filterObject.put("start", startTime);
          filterObject.put("end", endTime);
          if (!"".equals(firmId)) {
            filterObject.put("legalEntityId", firmId);
          }
          if (!"".equals(cofi)) {
            filterObject.put("mifidClassId", cofi);
          }
          jsonObject.put("filterObject", filterObject);
        } else if ("RTS27".equals(reportType)) {
          JSONObject filterObject = new JSONObject();
          filterObject.put("start", startTime);
          filterObject.put("end", endTime);
          jsonObject.put("filterObject", filterObject);
        }
      } catch (JSONException e) {
        e.printStackTrace();
        logger.error(e.getMessage());
      }
      token = this.getToken(username, password);
      JSONObject jobResponse = new JSONObject(
          this.edpApiClient.createJob(reportType, token, jsonObject.toString()));
      String jobId = this.getResponseJSONField("jobId", jobResponse);
      logger.info("Job submitted with job id: {}", jobId);

      return resultJson.put("jobId", jobId);
    } else {
      String errorInfo = "No files uploaded! Failed to run job";
      logger.info(errorInfo);
      return resultJson.put(ERROR_FLAG, errorInfo);
    }
  }

  /**
   * Execute Query Job Status job
   *
   * @param username   String
   * @param password   String
   * @param reportType String
   * @param jobId      String This is job id from previous RUN job's result
   * @return JSONObject {"jobId": "xxxxxxxxx", "jobMessage": {...}, "jobStatus": "yyyyyyyy"} on
   * successful; {"error": "error information"} on error
   * @throws IOException
   */
  public JSONObject executeQueryJob(String username, String password, String reportType,
      String jobId) throws IOException {
    if (!this.validateParamsForQueryJob(username, password, reportType, jobId)) {
      return new JSONObject().put(ERROR_FLAG, "Bad parameter(s)");
    }

    String token = this.getToken(username, password);

    return new JSONObject(this.edpApiClient.requestJobStatus(reportType, token, jobId));
  }

  /**
   * Execute Download Reports job. A .zip file will be download to specified local directory
   * <p>
   * Note: This function is via Job Id to download
   *
   * @param username      String
   * @param password      String
   * @param reportType    String
   * @param jobId         String This is job id from previous RUN job's result
   * @param localDir      String This is local directory to store the reports
   * @param encryptionKey String
   * @return JSONObject {"success": "true"} on successful; {"error": "error information"} on error
   * @throws IOException
   * @throws Exception
   */
  public JSONObject executeGetReportsJob(String username, String password, String reportType,
      String jobId, String localDir, String encryptionKey) throws IOException {
    JSONObject resultJson = new JSONObject();
    if (!this.validateParamsForGetReportsJob(username, password, reportType, jobId, localDir,
        encryptionKey)) {
      return resultJson.put(ERROR_FLAG, "Bad parameters(s)");
    }
    logger.info("Query job status");
    String[] reportS3Info = this.getS3InfoFromDoneJob(username, password, reportType, jobId);

    logger.info("Start downloading");
    String token = this.getToken(username, password);
    String bucketName = reportS3Info[0];
    String region = reportS3Info[1];
    String jobInfoFilePath = "";
    Pattern p = Pattern.compile("(.*/rept-srv-.*)/report/.*");
    Matcher m = p.matcher(reportS3Info[2]);
    if (m.find()) {
      jobInfoFilePath = m.group(1);
    } else {
      throw new IllegalArgumentException(
          "Can not parse jobInfoFilePath and jobId from report file path");
    }
    AWSCredentialsProvider credentialsProvider = this
        .getAWSCredentialsProviderWithEdpClouldCredentials(token, region, bucketName,
            jobInfoFilePath);
    S3FileTransfer.downloadFiles(localDir, reportS3Info[0], reportS3Info[1], reportS3Info[2], jobId,
        this.generateKey(username, jobId, encryptionKey), credentialsProvider);
    return resultJson.put("success", "true");
  }

  private boolean validateParamsForRunJob(String username, String password, String localDir,
      String encryptionKey, String startTime,
      String endTime, String email, String reportType) {
    if (!(this.validateUsername(username) && this.testLocalDir(localDir) && this
        .validateEmail(email))) {
      logger.info(MSG_FAIL_TO_RUN);
      return false;
    }
    if (StringUtils.isBlank(password)) {
      logger.info(MSG_CREDENTIALS_NULL);
      return false;
    }
    if (StringUtils.isBlank(encryptionKey)) {
      logger.info("Encryption key should not be blank");
      return false;
    }
    if (StringUtils.isNoneBlank(startTime) && StringUtils.isNoneBlank(endTime)) {
      if ("FITCA".equals(reportType)) {
        logger.info("-s and -e paramters are not allowed for report type FITCA");
        return false;
      }
      try {
        Date startDate = this.parseDate(startTime);
        Date endDate = this.parseDate(endTime);
        if (startDate.after(endDate)) {
          logger.info("Start date should not after end date");
          return false;
        }
      } catch (ParseException e) {
        logger.info("Invalid request date parameter");
        logger.error(e.toString());
        return false;
      }
    } else if (StringUtils.isBlank(startTime) || StringUtils.isBlank(endTime)) {
      /* allowed, default values will be applied */
      if ("RTS27".equals(reportType) || "RTS28".equals(reportType)) {
        logger.info("-s and -e paramters are required for report type %s ", reportType);
        return false;
      }
    }
    return true;
  }

  private boolean validateParamsForQueryJob(String username, String password, String reportType,
      String jobId) {
    if (!this.validateUsername(username)) {
      logger.info(MSG_FAIL_TO_RUN);
      return false;
    }
    if (StringUtils.isBlank(password)) {
      logger.info(MSG_CREDENTIALS_NULL);
      return false;
    }
    if (StringUtils.isBlank(reportType)) {
      logger.info("Report type should not be blank");
      return false;
    }
    if (StringUtils.isBlank(jobId)) {
      logger.info("Job ID should not be blank");
      return false;
    }

    return true;
  }

  private boolean validateParamsForGetReportsJob(String username, String password,
      String reportType, String jobId, String localDir, String encryptionKey) {
    if (!this.validateUsername(username)) {
      logger.info(MSG_FAIL_TO_RUN);
      return false;
    }
    if (StringUtils.isBlank(password)) {
      logger.info(MSG_CREDENTIALS_NULL);
      return false;
    }
    if (StringUtils.isBlank(encryptionKey)) {
      logger.info("Encryption key should not be blank");
      return false;
    }
    if (StringUtils.isBlank(reportType)) {
      logger.info("Report type should not be blank");
      return false;
    }
    if (StringUtils.isBlank(jobId)) {
      logger.info("Either Job ID or CFS report location is required");
      return false;
    }
    if (StringUtils.isBlank(localDir)) {
      logger.info("localDir should not be blank");
      return false;
    }

    return true;
  }

  private boolean validateUsername(String username) {
    if (StringUtils.isBlank(username)) {
      logger.info("Username should not be blank");
      return false;
    }
    if (!this.isEmailFormat(username)) {
      logger.info("Username is not in email format");
      return false;
    }
    return true;
  }

  private boolean validateEmail(String email) {
    if (StringUtils.isBlank(email)) {
      logger.info("Email should not be blank");
      return false;
    }
    if (!this.isEmailFormat(email)) {
      logger.info("Email is not in email format");
      return false;
    }
    return true;
  }

  private boolean isEmailFormat(String emailStr) {
    return Pattern.compile(".+@.+\\..+").matcher(emailStr).find();
  }

  private boolean testLocalDir(String dir) {
    if (StringUtils.isBlank(dir)) {
      logger.info("Directory parameter is blank");
      return false;
    }
    File f = new File(dir);
    if (f.exists()) {
      if (f.listFiles().length == 0) {
        logger.info("No file found under directory {}", dir);
        return false;
      }
      return true;
    } else {
      logger.info("Directory \"{}\" is not existing.", dir);
      return false;
    }
  }

  /**
   * @param date String Example: 2018-01-01
   * @throws ParseException
   */
  private Date parseDate(String date) throws ParseException {
    SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-dd");
    Date d = format.parse(date);
    if (!format.format(d).equals(date)) {
      throw new ParseException(date + " is not a valid date", -1);
    }
    return d;
  }

  /**
   * Get EDP authentication token
   *
   * @param username String
   * @param password String
   * @return String
   * @throws IOException
   * @throws
   */
  private String getToken(String username, String password) throws IOException {
    JSONObject response = new JSONObject(this.edpApiClient.requestToken(username, password, APPID));
    try { // check if request failed
      String error = response.getString(ERROR_FLAG);
      if (StringUtils.isNoneEmpty(error)) {
        throw new IllegalStateException(response.toString());
      }
    } catch (JSONException e) {/* intent blank */}
    logger.info("Got AAA token");
    return response.get("token_type").toString() + " " + response.get("access_token").toString();
  }

  /**
   * Get User's EDP Cloud Credentials for Access AWS
   *
   * @param token String EDP token
   * @return String[] {%accessKeyId%, %secrectAccessKey%, %sessionToken%}
   * @throws IOException
   */
  private String[] getEdpClouldCredentials(String token, String region, String bucketName,
      String jobInfoFilePath) throws IOException {
    JSONObject credentialsResponse =
        new JSONObject(this.edpApiClient
            .requestCredentials(token, EDP_ENDPOINT_4_CLOUD_CREDENTIAL, region, bucketName,
                jobInfoFilePath));
    logger.info("Got temporary cloud credentials");

    JSONObject credentials = (JSONObject) credentialsResponse.get("credentials");
    String accessKey = this.getResponseJSONField("accessKeyId", credentials);
    String screteKey = this.getResponseJSONField("secretKey", credentials);
    String sessionToken = this.getResponseJSONField("sessionToken", credentials);

    return new String[]{accessKey, screteKey, sessionToken};
  }

  private AWSCredentialsProvider getAWSCredentialsProviderWithEdpClouldCredentials(String username,
      String password, String region, String bucketName, String jobInfoFilePath)
      throws IOException {
    String token = this.getToken(username, password);

    String[] credentials = this.getEdpClouldCredentials(token, region, bucketName, jobInfoFilePath);

    return new AWSStaticCredentialsProvider(
        new BasicSessionCredentials(credentials[0], credentials[1], credentials[2]));
  }

  /**
   * @param token String This is the EDP authentication token
   * @return AWSCredentialsProvider
   * @throws IOException
   */
  private AWSCredentialsProvider getAWSCredentialsProviderWithEdpClouldCredentials(String token,
      String region, String bucketName, String jobInfoFilePath) throws IOException {

    String[] credentials = this.getEdpClouldCredentials(token, region, bucketName, jobInfoFilePath);

    return new AWSStaticCredentialsProvider(
        new BasicSessionCredentials(credentials[0], credentials[1], credentials[2]));
  }

  /**
   * Parse the job id from S3 file path
   *
   * @param filepath String This is the S3 file path and should be get by EDP API
   * @return String Job ID
   * @throws Exception
   */
  private String parseJobId(String filepath) {
    Matcher m = Pattern.compile(".*rept-srv-(.+)/.*").matcher(filepath);
    if (m.find()) {
      return m.group(1);
    } else {
      throw new IllegalArgumentException("Can not parse potential job id from file path");
    }
  }

  /**
   * @param token String
   * @return String[] [$s3Bucket, $s3Region, $s3Folder]
   * @throws IOException
   */
  private String[] getS3InfoFromUploadLocationRequest(String reportType, String token)
      throws IOException {
    logger.info("Got file upload location");
    JSONObject uploadLocation = new JSONObject(
        this.edpApiClient.requestUploadLocation(reportType, token));

    String s3Bucket = this.getResponseJSONField("bucket", uploadLocation);
    if (StringUtils.isBlank(s3Bucket)) {
      throw new IllegalArgumentException(
          "No S3 bucket name from UploadLocation request's reponse.");
    }
    String s3Folder = this.getResponseJSONField("prefix", uploadLocation);
    if (StringUtils.isBlank(s3Folder)) {
      throw new IllegalArgumentException(
          "No S3 bucket folder info from UploadLocation request's reponse.");
    }
    String s3Region = this.getResponseJSONField("region", uploadLocation);
    if (StringUtils.isBlank(s3Region)) {
      throw new IllegalArgumentException(
          "No S3 bucket region info from UploadLocation request's reponse.");
    }
    return new String[]{s3Bucket, s3Region, s3Folder};
  }


  /**
   * @param username   String
   * @param password   String
   * @param reportType String
   * @param jobId      String
   * @return String[] [%s3Bucket%, %s3Region%, %s3Folder%]
   * @throws IOException
   */
  private String[] getS3InfoFromDoneJob(String username, String password, String reportType,
      String jobId) throws IOException {
    JSONObject jobStatusResponse = this.executeQueryJob(username, password, reportType, jobId);
    String jobState = this.getResponseJSONField("jobState", jobStatusResponse);
    if (StringUtils.isBlank(jobState)) {
      throw new IllegalArgumentException("No jobState in job status message.");
    }
    if (!"completed".equalsIgnoreCase(jobState)) {
      logger.info("Job is not completed yet. Current status is {}", jobState);
      throw new IllegalArgumentException("Job is not completed yet.");
    }

    JSONObject msgJson = jobStatusResponse.getJSONObject("jobMessage");
    if (msgJson == null) {
      throw new IllegalArgumentException("No job message in job status message.");
    }
    String filePath = this.getResponseJSONField("filepath", msgJson);
    String[] pathInfo;
    if (StringUtils.isBlank(filePath)) { // S3 bucket key of report file
      throw new IllegalArgumentException("No report S3 file key in job status message.");
    } else {
      if (filePath.startsWith("/file-store")) { // get report location from CFS API
        String token = this.getToken(username, password);
        pathInfo = this.parseReportFilePath(this.getS3PathFromCFS(token, filePath));
      } else if (filePath.startsWith(
          "https://s3.amazonaws.com")) { // report location is immediately available and no need to access CFS
        pathInfo = this.parseReportFilePath(filePath);
      } else {
        throw new IllegalArgumentException(
            "Wrong format of report file path in job status message.");
      }
    }
    String s3Region = this.getResponseJSONField("region", msgJson);
    if (StringUtils.isBlank(s3Region)) {
      throw new IllegalArgumentException("No S3 region info in job status message.");
    }

    return new String[]{pathInfo[0], s3Region, pathInfo[1]};
  }

  private String[] parseReportFilePath(String path) {
    Pattern p = Pattern.compile("https://s3.amazonaws.com/([^/]*)/(.*/rept-srv-(.*)/report/.*)");
    Matcher m = p.matcher(path);
    if (m.find()) {
      String s3Bucket = m.group(1);
      String s3Folder = m.group(2);
      String jobId = m.group(3);
      return new String[]{s3Bucket, s3Folder, jobId};
    } else {
      throw new IllegalArgumentException("Can not parse CFS report location string");
    }
  }

  private String getS3PathFromCFS(String token, String cfsPath) throws IOException {
    logger.info("Request CFS report information");
    JSONObject cfsResponse = new JSONObject(
        this.edpApiClient.requestCFSReportLocation(token, cfsPath));
    String locationUrl = cfsResponse.getJSONObject("storageLocation").getString("url");
    if (StringUtils.isBlank(locationUrl)) {
      throw new IllegalArgumentException("Report location from CFS is blank");
    }

    return locationUrl;
  }

  /**
   * Encrypt the user original encrypt key for protection purpose
   *
   * @param userId  String This value should be an email address
   * @param jobId   String
   * @param keySeed String This is the user original key
   * @return String encrypted encrypt key with length 32
   * @throws Exception
   */
  private String generateKey(String userId, String jobId, String keySeed) {
    int atPos = userId.indexOf('@');
    String username = userId.substring(0, atPos);
    String userCompany = userId.substring(atPos + 1);

    String plainText = username + "__" + keySeed + "__" + userCompany + "__" + jobId;

    return DigestUtils.sha256Hex(plainText).substring(0, 32);
  }

  private String getResponseJSONField(String fieldName, JSONObject json) {
    try {
      return json.get(fieldName).toString();
    } catch (JSONException e) {
      logger.error("No {} filed found in response: {}", fieldName, json);
      throw e;
    }
  }
}
