/*
 *  Copyright (c) 2019 by Refinitiv. All rights reserved.
 *
 *  No portion of this software in any form may be used or
 *  reproduced in any manner without written consent from
 *  Refinitiv
 */
package com.refinitiv.analytics.reportingservice;

import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Main class for MiFID Reporting Application
 */
public class Main {

  private static final Logger logger = LoggerFactory.getLogger(Main.class);

  public static void main(String[] args) {
    // 1. Java Version needs to be >= 8u161(1.8.0_161), otherwise a warning will be given.
    validateJavaVersion();

    // 2. parse the command line arguments
    CmdLineHandler cmdLine = new CmdLineHandler();
    if (!cmdLine.parseArgs(args)) {
      return;
    }

    // 3. run respective job according to command line argument
    JobHandler jobHandler = new JobHandler();
    jobHandler.execute(cmdLine);
  }

  /**
   * Probe Java version is not less than 8u161
   * <p>
   * With Java version under 8u161, Java Cryptography Extension (JCE) jurisdiction policy limits the
   * maximum encrypt key length to 128 bits. While AWS S3 SDK requires 256 bits
   * <p>
   * Sample Java version String: "11.0.1+13", "10.0.99", "9.0.99", "1.8.0_191-b12", "1.7.0_201-b11",
   * "1.6.0_211-b11"
   */
  private static boolean validateJavaVersion() {
    String version = Runtime.class.getPackage().getImplementationVersion();
    Pattern p = Pattern.compile("^(\\d+)\\.(\\d+)\\.(\\d+)(_(\\d{3}).*)?");
    Matcher m = p.matcher(version);
    if (m.find()) {
      try {
        int major = Integer.parseInt(m.group(1));
        if (major >= 9) {
          return true; // true, major version grater than 8
        } else if (major == 1) {
          major = Integer.parseInt(m.group(2));
          int minor = Integer.parseInt(m.group(5));
          if (major == 8 && minor >= 161) {
            return true;  // version not less than 8u161
          }
          logger.info("Warning: Upgrade to Java Version 8u161 or above, current version is {}."
                  + " With Java version under 8u161, Java Cryptography Extension (JCE) jurisdiction policy limits"
                  + " the maximum encrypt key length to 128 bits. While AWS S3 SDK requires 256 bits.",
              version);
          return false; // version less than 8u161
        }
      } catch (NumberFormatException e) {/* intent blank */}
    }
    logger.info("Warning: Not able to detect java version. It should be not less than 8u161");
    return true; // Not able to detect java version
  }
}
