/*
 *  Copyright (c) 2019 by Refinitiv. All rights reserved.
 *
 *  No portion of this software in any form may be used or
 *  reproduced in any manner without written consent from
 *  Refinitiv
 */
package com.refinitiv.analytics.reportingservice.utils;

import com.amazonaws.ClientConfiguration;
import com.amazonaws.Protocol;
import com.amazonaws.auth.AWSCredentialsProvider;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3EncryptionClientBuilder;
import com.amazonaws.services.s3.model.EncryptionMaterials;
import com.amazonaws.services.s3.model.GetObjectRequest;
import com.amazonaws.services.s3.model.S3Object;
import com.amazonaws.services.s3.model.S3ObjectInputStream;
import com.amazonaws.services.s3.model.StaticEncryptionMaterialsProvider;
import com.refinitiv.analytics.reportingservice.config.SysConfig;
import java.io.File;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.regex.Pattern;
import javax.crypto.spec.SecretKeySpec;
import org.aeonbits.owner.ConfigFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class S3FileTransfer {

  private static final Logger logger = LoggerFactory.getLogger(S3FileTransfer.class);
  private static final SysConfig config =
      ConfigFactory.create(SysConfig.class);

  private S3FileTransfer() {
  }

  public static ClientConfiguration setProxy() {
    String proxyIp = config.getNetworkProxyIp();
    int proxyPort = config.getNetworkProxyPort();
    ClientConfiguration clientConfiguration = null;
    if (proxyIp != null && proxyPort != -1 && !"".equals(proxyIp)) {
      clientConfiguration = new ClientConfiguration();
      clientConfiguration.setProtocol(Protocol.HTTP);
      //setup proxy connection:
      clientConfiguration.setProxyHost(proxyIp);
      clientConfiguration.setProxyPort(proxyPort);
    }
    return clientConfiguration;
  }

  public static boolean uploadFile(File file, String s3Bucket, String s3Region,
      String s3Folder, String encryptKey, AWSCredentialsProvider credentialsProvider) {

    EncryptionMaterials encryptionMaterials =
        new EncryptionMaterials(new SecretKeySpec(encryptKey.getBytes(), "AES"));
    AmazonS3 encryptionClient = AmazonS3EncryptionClientBuilder.standard()
        .withCredentials(credentialsProvider)
        .withRegion(s3Region)
        .withEncryptionMaterials(new StaticEncryptionMaterialsProvider(encryptionMaterials))
        .withClientConfiguration(setProxy())
        .build();

    if (file.isFile()) {
      String fileName = file.getName();
      String targetFile = s3Folder + "/" + fileName;
      encryptionClient.putObject(s3Bucket, targetFile, file);
      logger.info("{} uploaded!", fileName);
      return true;
    }
    return false;
  }

  /**
   * Skip files from uploading: - name contains whitespace character - not csv file
   */
  public static boolean shouldSkip(String fileName) {
    if (Pattern.compile("\\s+").matcher(fileName).find()) {
      logger.info("{} skipped due to whitespace char", fileName);
      return true;
    }
    if (!fileName.equals("ColumnsMapping") && !Pattern.compile(".csv$").matcher(fileName).find()) {
      logger.info("{} skipped due to it is not csv", fileName);
      return true;
    }
    return false;
  }

  public static boolean downloadFiles(String localDir, String s3Bucket, String s3Region,
      String s3Path,
      String jobId, String encryptKey, AWSCredentialsProvider credentialsProvider)
      throws IOException {

    EncryptionMaterials encryptionMaterials =
        new EncryptionMaterials(new SecretKeySpec(encryptKey.getBytes(), "AES"));
    AmazonS3 encryptionClient = AmazonS3EncryptionClientBuilder.standard()
        .withCredentials(credentialsProvider)
        .withRegion(s3Region)
        .withEncryptionMaterials(new StaticEncryptionMaterialsProvider(encryptionMaterials))
        .withClientConfiguration(setProxy())
        .build();

    if (Paths.get(localDir).toFile().isDirectory()) {
      File destDir = new File(localDir);
      if (!destDir.exists()) {
        Files.createDirectory(FileSystems.getDefault().getPath(destDir.getAbsolutePath()));
      }
    }

    String locaReportFullPath = localDir + File.separator + "Reports-" + jobId + ".zip";
    File localZip = new File(locaReportFullPath);
    logger.info("Start downloading report...");
    S3Object s3Object = null;
    S3ObjectInputStream s3InputStream = null;
    BufferedOutputStream outputStream = null;

    try {
      s3Object = encryptionClient.getObject(new GetObjectRequest(s3Bucket, s3Path));
      if (null == s3Object) {
        logger.error(String.format("Failed to get S3 Object from %s/%s", s3Bucket, s3Path));
        return false;
      }

      s3InputStream = s3Object.getObjectContent();
      if (null == s3InputStream) {
        logger.error(String.format("Failed to get S3 Content Stream from %s/%s", s3Bucket, s3Path));
        return false;
      }

      outputStream = new BufferedOutputStream(new FileOutputStream(localZip));
      byte[] buffer = new byte[10240];

      int bytesRead;
      while ((bytesRead = s3InputStream.read(buffer)) > -1) {
        outputStream.write(buffer, 0, bytesRead);
      }
    } catch (Exception e) {
      if (e.getMessage().contains("Unable to decrypt symmetric key")) {
        logger.info(
            "Failed to download with a mismatched key, please use the same key to the same jobId");
      }

      throw e;
    } finally {
      if (null != outputStream) {
        outputStream.close();
      }

      if (null != s3InputStream) {
        while (s3InputStream.read() > -1) {
          //still have something buffered, clean buffer of the S3 stream before close
        }
        s3InputStream.close();
      }

      if (null != s3Object) {
        s3Object.close();
      }
    }
    return true;
  }
}
